"""The pytest tool invokes pytest based tests."""
import argparse
import logging

from menhir.tool import TEST, Tool
from menhir.tool_utils import (
    OK,
    FAIL,
    NOTHING_TO_DO,
    changed_state,
    has_self_or_dependent_changes,
    package_script,
    tool_env,
    working_dir,
)

log = logging.getLogger(__name__)


def tool():
    return PyTest()


class PyTest(Tool):

    def dir_info(tool, path, info):
        from os.path import exists, join
        has_pytest = None
        path = join(path, 'setup.cfg')
        if exists(path):
            with open(path, "r") as file:
                data = file.read()
                has_pytest = 'pytest' in data
        return {
            'project_recognised': has_pytest,
            'can_run': has_pytest,
        }

    def dependencies(tool, path):
        return []

    def arg_parser(tool, **kwargs):
        return parser(**kwargs)

    def execute_tool(tool, path, info, args,):
        """Execute a build phase."""
        import subprocess

        log.info('pytest: %s %s %s', path, info, args)

        with working_dir(path):
            changed = changed_state(info)

            if has_self_or_dependent_changes(changed):
                log.info('Running pytest in %s', path)
                project_name = info['project-name']
                env = tool_env()
                env['MENHIR_PROJECT'] = project_name
                with package_script("/tools/pytest/test.sh") as f:
                    res = subprocess.call([f.name] + args.args, env=env)
                    if res:
                        return FAIL
                    return OK
            else:
                return NOTHING_TO_DO

    def build_jobs(tool, info):
        return [TEST]


def parser(**kwargs):
    parser = argparse.ArgumentParser(
        description="Invoke pytest based tests",
        **kwargs
    )
    parser.add_argument('args', nargs='*', help='pytest arguments')
    return parser



"""The git tool provides tasks for working with a git reopsitory.

The ``install-hooks`` task installs local git hooks.  Hooks are
assumed to be in a top-level ``git-hooks`` directory.  This can be
configured with the ``.git.hook-dir`` configuration value.

The ``with-stashed-changes`` task runs whatever command it it is
passed with all unstaged commits stashed.

"""
import argparse
import logging
import stat

from menhir.tool import CHECKOUT, Tool
from menhir.tool_utils import OK, FAIL
from menhir.utils import multi, method

log = logging.getLogger(__name__)

RWXR_XR_X = (
    stat.S_IRUSR |
    stat.S_IWUSR |
    stat.S_IXUSR |
    stat.S_IRGRP |
    stat.S_IXGRP |
    stat.S_IROTH |
    stat.S_IXOTH
)


def tool():
    return Git()


class Git(Tool):

    def dir_info(tool, path, info):
        from os.path import isdir, join
        path = join(path, '.git')
        has_git = isdir(path)
        return {
            'project_recognised': has_git,
            'can_run': has_git,
        }

    def dependencies(tool, path):
        return []

    def arg_parser(tool, **kwargs):
        return parser(**kwargs)

    def execute_tool(tool, path, info, args):
        """Execute git task."""
        return task(path, info, args)

    def build_jobs(tool, info):
        return [CHECKOUT]


def parser(**kwargs):
    parser = argparse.ArgumentParser(
        description="git utilities",
        **kwargs
    )
    parsers = parser.add_subparsers(help="git tasks", dest='phase')

    p = parsers.add_parser(
        'install-hooks',
        help='Install local git hooks'
    )
    p.add_argument('remainder', nargs=argparse.REMAINDER)

    p = parsers.add_parser(
        'with-stashed-changes',
        help='Execute a command with all unstaged changes stashed'
    )
    p.add_argument('cmd', nargs='*', help='command to execute')
    p.add_argument('remainder', nargs=argparse.REMAINDER)

    return parser


@multi
def task(path, info, args):
    """Execute a task."""
    return args.phase


@method(task, 'install-hooks')
def install_hooks(path, info, args):
    """Install local git hooks from a project directory."""
    from os import chmod, listdir
    from os.path import isdir, isfile, join, normpath
    from shutil import copy

    log.debug('git install-hooks')
    hook_dir_name = info.get('git', {}).get('hook-dir', "git-hooks")
    hook_dir = join(path, hook_dir_name)
    if not isdir(hook_dir):
        log.error('No git hooks directory %s', normpath(hook_dir))
        return FAIL

    target_dir = join(path, ".git", "hooks")
    for f in listdir(hook_dir):
        log.debug('Found potential hook file %s', f)
        src = join(hook_dir, f)
        if isfile(src):
            log.debug('Install git hook %s', f)
            target = join(target_dir, f)
            copy(src, target)
            chmod(target, RWXR_XR_X)

    return OK


@method(task, 'with-stashed-changes')
def with_stashed_changes(path, info, args):
    from menhir import gitutils
    from menhir.tool_utils import call, working_dir

    repo = gitutils.repo()
    with working_dir(path):
        with gitutils.staged_as_committed(repo):
            with gitutils.stashed_changes(repo):
                return call(args.cmd)



"""Tool interface for tool implementations.

A tool is implemented as an opaque class.

"""
import logging

from dialogue.multi_method import method, multi

log = logging.getLogger(__name__)


CHECKOUT = 'checkout'
CHECKOUT_STATE = 'checkout_state'
DEPENDENCIES = 'dependencies'
TEST = 'test'
LINT = 'lint'
PACKAGE = 'package'
DEPLOY = 'deploy'


def missing_impl(tool, f):
    raise Exception(
        'incorrect tool implementation',
        'Tool implementation %s is missing a "%s" implementation' % (
            type(tool),
            f,
        )
    )


class Tool(object):
    """The Tool interface implemented by each tool."""
    def __str__(self):
        return "<%s>" % type(self).__name__

    def __repr__(self):
        return "<%s>" % type(self).__name__

    def name(self):
        """Return the name of the tool on the command line."""
        return type(self).__name__.lower()

    def dir_info(tool, path, info):
        """Return info about using the tool in the specified path."""
        missing_impl(tool, Tool.dir_info.__name__)

    def dependencies(tool, path):
        """Return a list of dependency prefixes for the project at path."""
        missing_impl(tool, Tool.dependencies.__name__)

    def configure_parser(tool, parser):
        """Configure arg parser for the tool options and arguments."""
        missing_impl(tool, Tool.add_arg_parser.__name__)

    def execute_tool(tool, path, info, args):
        """Execute a build phase."""
        missing_impl(tool, Tool.execute_tool.__name__)

    def build_jobs(tool, info):
        """Return build job specs."""
        missing_impl(tool, Tool.build_jobs.__name__)

    def traverse_multiproject(tool, info):
        """Predicate for multiproject traversal.

        Allow tools that do not expect traversal (eg. to do something only in
        the project root).

        """
        return True


def load_tool(tool_name):
    """Load the tool implementation for the given tool name."""
    builtin = 'menhir.tools.%s' % tool_name
    tool, e1 = require_tool(builtin)
    if not tool:
        tool, e2 = require_tool(tool_name)
    if not tool:
        raise Exception(
            'configuration_error',
            'Tool not found: "%s".  import %s gave %s.  import %s gave %s' % (
                tool_name,
                builtin, e1,
                tool_name, e2
            ))
    return tool


def require_tool(tool_path):
    from importlib import import_module
    try:
        m = import_module(tool_path)
        return m.tool(), None
    except Exception as e:
        log.debug('Failed to import: %s, "%s"', tool_path, e)
        return None, e


def default_tools():
    """Return a list of the default tool names."""
    from menhir.tools import all_tools
    return all_tools


@multi
def build_impl(tool, task, build_env, info):
    """Multi-method to return invocations for tool to run task on buildenv."""
    return tool.name(), task, build_env


@method(build_impl)
def build_impl_default(tool, task, build_env, info):
    return []



all_tools = [
    'codecov',
    'docker',
    'dynamodb',
    'exec',
    'git',
    'multiproject',
    'pprint',
    'pyenv',
    'pylama',
    'pytest',
    'setup_py',
    'terraform',
    'write_state',
]



"""Multi-project build tool invoker.

A build tool to invoke a tool on a group of co-dependent projects.

Menhir `multiproject` builds a graph of the dependencies between
projects in a source repository.  It does this with the help of the
build tools.

The :doc:`setup_py` tool contributes dependencies based on
``file://..`` links in ``requirements.txt``.


Configuration
-------------

The `multiproject` key in :file:`menhir-config.yaml` is used to
configure the projects, and how :command:`menhir` discovers projects.

When ``project_graph``, ``root_projects`` and ``project_config`` are
specified in configuration, no project discovery is done.

To output discovered configuration in a form that can be saved in
:file:`menhir-config.yaml`, use the command::

    menhir multiproject --no-traverse pprint

The available keys:

``tools``
    a list of available tools to use in project discovery.
    The multiproject tool itself is always used in discovery,
    and recognises directories containing ``.menhir_project``
    or ``menhir-config.yaml`` files.

    If ``menhir-config.yaml`` contains a ``depends_on`` key, with
    a list of project relative (sub-)project paths, these are added
    to the projects dependencies.


``ignore``
    a list of directory names to ignore (all hidden directories,
    starting with a period are ignored by default, as are
    ``__pycache__`` directories).


``project_graph``
    a dict specifying dependency edges between project paths.

``root_projects``
    a list of project paths at the root of the dependency graph.

``project_config``
    a dict specify project configuration.  At present the only
    supported key is ``tools``, specify the tools the project uses.


Default configuration
---------------------

The multiproject tool propagates configuration down the source tree.
It also reads a default configuration, inherited by projects lower
down the source tree, from a :file:`menhir-defaults.yaml` file with
the same format as :file:`menhir-config.yaml`.


Git Status
----------

When run with the :command:`git_status` tool, the changed state of
each project, including dependent and dependee changes, is set, and
this information passed to the build tools that
:command:`multiproject` invokes.

Each tool provides it's own specific behaviour, depending on the the
status of each project, and on whether the changed information is
available or not.

"""
from __future__ import print_function  # NOQA

import argparse
import logging
import os.path

from menhir.fileutils import load_yaml
from menhir.tool import Tool
from menhir.tool_utils import FAIL, NOTHING_TO_DO, OK, call_tool_chain
from menhir.utils import getd

log = logging.getLogger(__name__)


def tool():
    return MultiProject()


class MultiProject(Tool):

    def dir_info(tool, path, info):
        has_marker = os.path.exists(os.path.join(path, ".menhir-project"))
        has_config = os.path.exists(os.path.join(path, "menhir-config.yaml"))
        return {
            'project_recognised': has_marker or has_config,
            'can_run': True,
        }

    def dependencies(tool, path):
        config = load_yaml(os.path.join(path, 'menhir-config.yaml')) or {}
        if 'souvenir' in path:
            print('config', config)
        return config.get('depends_on', [])

    def arg_parser(tool, **kwargs):
        return parser(**kwargs)

    def execute_tool(tool, path, info, args):
        return call_tool(path, info, args)

    def build_jobs(tool, info):
        return []


def call_tool(path, info, args):
    from os.path import relpath
    from menhir.tool import load_tool

    if not args.sub_tool:
        return NOTHING_TO_DO

    root = args.root or info.get('menhir', {}).get('root')
    root = relpath(root)
    info = config_with_projects(info, root)

    config = info['multiproject']
    state = info['multiconfig']

    project_graph = config['project_graph']
    project_meta = config['project_meta']
    root_projects = config['root_projects']
    project_infos = state['project_infos']

    git_status = info.get('git_status', {})
    files = (
        git_status.get('uncommited') or
        git_status.get('changed', {}).get('all')
    )
    if files:
        update_for_changed_files(
            project_graph, project_infos, root_projects, root, files
        )

    no_traverse = args.no_traverse
    tool_name = args.sub_tool
    tool = load_tool(tool_name)
    no_traverse = no_traverse or not tool.traverse_multiproject(info)

    parser = tool.arg_parser(prog=tool_name)
    args = parser.parse_args(args.arg)

    if no_traverse:
        print('project_meta', project_meta)
        meta = project_meta.get(root, {})
        res = apply_tool_to_path(path, info, meta, tool, args)
    else:
        res = apply_tool(
            project_graph,
            project_infos,
            root_projects,
            project_meta,
            tool,
            args
        )

    if res.get('fail'):
        log.error('Build tool "%s" failed', tool_name)

    log.debug('res: %s', res)
    return res


def parser(**kwargs):
    """Arg parser for the tool options and arguments."""
    parser = argparse.ArgumentParser(
        description="Multi-project build tool invoker.",
        **kwargs
    )

    parser.add_argument(
        "--root",
        help="Specify the root of the project tree.",
    )

    parser.add_argument(
        "--no-traverse",
        default=False,
        action='store_const',
        const=True,
        help="""Only run the given command once.
The  project graph is not traversed.""",
    )

    parser.add_argument(
        "sub_tool",
        nargs="?",
        help='Tool to invoke.',
    )

    parser.add_argument(
        "arg",
        nargs=argparse.REMAINDER,
        help='Tool arguments',
    )

    return parser


def config_with_projects(info, rel_root):
    from menhir import gitutils
    rel_root = rel_root or gitutils.find_root()
    config = getd(info, 'multiproject')
    infos = getd(info, 'multiconfig')
    info['multiconfig'] = infos
    if not config.get('project_graph'):
        info = describe_repository(info, rel_root)
        config = getd(info, 'multiproject')
        infos = getd(info, 'multiconfig')

    infos['root'] = rel_root
    return info


def describe_repository(root_info, rel_root):
    """Parse the root menhir configuration."""
    from menhir.fileutils import directory_graph
    from menhir.graph import root_nodes
    from menhir.tool import default_tools, load_tool

    tools = root_info.get('menhir', {}).get('tools', default_tools())
    tool_impls = list(map(load_tool, tools))  # dict(zip(tools, ))

    config = getd(root_info, 'multiproject')

    ignore_dirs = config.get('ignore', [])
    log.debug('Ignoring: %s', ignore_dirs)

    # build the dependency graph
    fs_graph = directory_graph(rel_root, ignore_dirs)
    project_meta = config.get('project_meta')

    project_meta, infos = project_infos(
        fs_graph, rel_root, tool_impls, project_meta
    )

    graph = config.get('project_graph', project_graph(project_meta))
    roots = config.get('root_projects', root_nodes(graph, infos))

    def is_child(x, y):
        from os.path import abspath, normpath
        return normpath(abspath(y)).startswith(normpath(abspath(x)))

    roots = list(filter(lambda x: is_child('.', x), roots))

    config['project_meta'] = project_meta
    config['project_graph'] = graph
    config['root_projects'] = roots
    root_info['multiconfig'] = root_info.get('multiconfig', {})
    root_info['multiconfig']['project_infos'] = infos
    return root_info


def project_infos(graph, root, tools, project_meta):
    """Return a map of project root path to project info dict."""
    from os.path import abspath, basename
    from menhir.graph import dfs_visit
    from menhir.config import project_config

    create_meta = not bool(project_meta)
    project_meta = project_meta or {}

    infos = {}

    def calc_config(defaults, path):
        from copy import deepcopy
        from menhir.utils import deep_merge
        defaults = deep_merge(defaults, defaults_config(path))
        config = deepcopy(defaults)
        config = deep_merge(config, project_config(dir=path))
        config['multiproject'] = {}
        return defaults, config

    def pre_f(defaults, path):
        defaults, config = calc_config(defaults, path)
        if create_meta:
            used_tools = list(filter(
                lambda t: t.dir_info(path, config).get('project_recognised'),
                tools
            ))
            used = list(map(lambda x: x.name(), used_tools))
            if used:
                project_meta[path] = project_meta.get(path, {})
                project_meta[path]['tools'] = used
        else:
            used = project_meta.get(path, {}).get('tools')

        if used:
            infos[path] = config
            infos[path]['project-name'] = (
                config.get('project-name') or
                basename(abspath(path))
            )

        return defaults

    def post_f(defaults, path, child_defaults):
        return defaults

    defaults = {}
    dfs_visit(graph, root, defaults, pre_f, post_f,)

    return project_meta, infos


def defaults_config(dir_path):
    """Return the defaults configuration in dir_path."""
    from os.path import join
    return load_yaml(join(dir_path, 'menhir-defaults.yaml')) or {}


def project_graph(roots):
    """Return a project dependency graph."""
    from functools import reduce
    from menhir.tool import load_tool

    edges = {
        root: reduce(
            lambda x, y: x.union(y or set()),
            [
                set(load_tool(tool).dependencies(root))
                for tool in set(info['tools']).union({'multiproject'})
            ],
            set()
        )
        for root, info in roots.items()
        }

    return {k: list(v) for k, v in edges.items()}


def apply_tool(dependencies, infos, roots, project_meta, tool, args):
    """Apply tool to each project.

    Projects are visited in dependency order.  A full traversal is
    made.  It is up to each tool to decide whether to run, based on
    the changed state in the project info.
    """
    from functools import reduce
    from menhir.graph import dfs_visit

    log.debug('apply_tool')

    def pre_f(state, path):
        return state

    def merge_state(x, y):
        log.debug('merge_state: %s %s', x, y)
        x['failed'] = x.get('failed') or y.get('failed')
        x['seen'] = x['seen'].union(y['seen'])
        return x

    def post_f(state, path, child_states):
        log.debug('apply_tool post_f %s %s', path, state)
        info = infos[path]
        meta = project_meta[path]
        if not state.get('failed') and path not in state['seen']:
            res = apply_tool_to_path(path, info, meta, tool, args)
            if res['status'] == 'fail':
                log.warning('%s failed in %s', tool, path)
                state['failed'] = True

        seen = state['seen']
        seen.add(path)
        state['seen'] = seen
        return state

    state = {'seen': set()}
    child_states = [
        dfs_visit(
            dependencies,
            root,
            state,
            pre_f,
            post_f,
            child_state_f=lambda x: x  # shared state
        )
        for root in roots
    ]

    log.debug('child_states: %s', child_states)
    state = reduce(merge_state, child_states, state)
    log.debug('merged state: %s', state)
    if state.get('failed'):
        return FAIL
    return OK


def apply_tool_to_path(path, info, meta, tool, args):
    log.debug('Executing %s in %s', tool.name(), path)
    if not tool.dir_info(path, info).get('can_run'):
        log.debug('Tool %s not runnable in %s', tool, path)
        return NOTHING_TO_DO
    res = tool.execute_tool(path, info, args)
    log.debug('Execute tool %s in %s returns %s', tool.name(), path, res)
    if res['status'] != 'ok':
        log.debug('Tool %s in %s result %s', tool.name(), path, res)
        res['tool'] = tool.name()
        res['args'] = args
        res['path'] = path
        return res

    remainder = getattr(args, 'remainder', None)
    log.debug('apply_tool_to_path remainder: %s', remainder)
    if remainder:
        return call_tool_chain(path, info, remainder)
    else:
        return res


def apply_cmd_to_path(path, info, cmd):
    import subprocess
    from menhir.tool_utils import working_dir
    log.info('Executing "%s" in %s', " ".join(cmd), path)

    with working_dir(path):
        rc = subprocess.call(cmd)
    log.debug('Execute cmd %s in %s returns %s', cmd, path, rc)

    if rc != 0:
        log.debug('Cmd %s in %s failed %s', cmd, path, rc)
        res = FAIL.copy()
        res.update({
            'cmd': cmd,
            'path': path
        })
        return res

    return OK


def update_for_changed_files(
        dependencies, project_infos, root_projects, prefix, files
):
    """Update project infos with changed status, given changed files.

    Updates project_infos with a `changed` map containing `self`,
    `dependees` and `dependents` fields, based on the list of changed
    files.  The `changed` key is present, even if none of it's keys
    are, so tooling knows that change information is present, and
    can modify it's behaviour accordingly.

    `prefix` allows the specification of a path prefix, to add to each
    of the project dirs, before determining if they contain any of the
    files.

    Returns a list of changed projects.

    """
    changed_projects = projects_with_files(project_infos, files, prefix)
    mark_changed(project_infos, changed_projects)
    propagate_changed(dependencies, project_infos, root_projects)
    return changed_projects


def projects_with_files(roots, files, root_prefix):
    """Filter roots that contain any of files paths.

    The roots are each prefixed with root_prefix, before comparison.
    """
    from os.path import join, normpath
    log.debug('projects_with_files:')
    log.debug(' roots: %s', roots)
    log.debug(' files: %s', files)
    roots = set(roots)
    changed_roots = set()

    for path in files:
        if len(roots):
            for root in roots.copy():
                prefix = normpath(join(root_prefix, root)) if root_prefix \
                         else root
                log.debug('prefix: %s, path: %s', prefix, path)
                if path.startswith(prefix) or (
                        prefix == '.' and not path.startswith('..')
                ):
                    log.debug('adding changed root %s', root)
                    changed_roots.add(root)
                    roots.remove(root)
        else:
            break
    return changed_roots


CHANGED = 'changed'
SELF = 'self'
DEPENDENTS = 'dependents'
DEPENDEES = 'dependees'


def mark_changed(projects, changed_projects):
    """Add a `changed` dict to each project, with `self` key boolean value."""
    for project in projects:
        projects[project][CHANGED] = {SELF: project in changed_projects}


def propagate_changed(dependencies, infos, root_projects):
    """Propagate changed info to dependees and dependents.

    `changed` nodes propagate `dependees` down the graph.
    `changed` nodes propagate `dependents` up the graph.
    """
    from menhir.graph import dfs_visit

    def pre_f(state, node):
        info = infos[node]
        if state.get(DEPENDEES):
            info[CHANGED][DEPENDEES] = True
        if info[CHANGED].get(SELF):
            state[DEPENDEES] = True
        log.debug('info pre: %s', info)
        return state

    def post_f(state, node, child_states):
        from copy import copy
        info = infos[node]
        for child_state in child_states:
            if (
                    child_state[CHANGED].get(DEPENDENTS) or
                    child_state[CHANGED].get(SELF)
            ):
                info[CHANGED][DEPENDENTS] = True
        log.debug('info post: %s', info)
        return copy(info)

    for root in root_projects:
        dfs_visit(dependencies, root, {}, pre_f, post_f)



"""Pyenv based tool for menhir.

The ``pyenv`` tool provides the ``virtualenv`` target to create a
virtualenv for each of the repository's projects.  The virtualenv is
loaded with all the project's dependencies as specified in
``requirements*.txt``.

The version of python to be used can be specified in the menhir configuration
using the `python-version` key, under the `pyenv` key.

"""
from __future__ import absolute_import

import argparse
import logging
import subprocess
from os import getenv
from os.path import expanduser, isdir, join

from git import Repo

from menhir.fileutils import slurp
from menhir.tool import DEPENDENCIES, Tool
from menhir.tool_utils import (
    FAIL,
    OK,
    changed_state,
    has_self_or_dependent_changes,
    package_script,
    tool_env,
    working_dir
)
from menhir.utils import method, multi

log = logging.getLogger(__name__)


def tool():
    return Pyenv()


class Pyenv(Tool):

    def dir_info(tool, path, info):
        from glob import glob
        from os.path import exists, join
        setup_py = join(path, 'setup.py')
        requirements = join(path, 'requirements*.txt')
        files = glob(requirements)
        has_requirements = exists(setup_py) or bool(files)
        return {
            'project_recognised': has_requirements,
            'can_run': has_requirements,
        }

    def dependencies(tool, path):
        return []

    def arg_parser(tool, **kwargs):
        return parser(**kwargs)

    def execute_tool(tool, path, info, args,):
        """Execute a build phase."""
        return task(path, info, args)

    def build_jobs(tool, info):
        return [DEPENDENCIES]


def parser(**kwargs):
    parser = argparse.ArgumentParser(
        description="Manage project specific pyenv virtualenvs",
        **kwargs
    )
    parsers = parser.add_subparsers(help="pyenv commands", dest='phase')
    p = parsers.add_parser(
        'virtualenv',
        help='Build a pyenv virtualenv with the projects dependencies'
    )
    p.add_argument('remainder', nargs=argparse.REMAINDER)

    p = parsers.add_parser(
        'install-pyenv',
        help='Install pyenv'
    )
    p.add_argument('remainder', nargs=argparse.REMAINDER)

    p = parsers.add_parser(
        'install-virtualenv',
        help='Install pyenv-virtualenv (assumes pyenv installed)'
    )
    p.add_argument('remainder', nargs=argparse.REMAINDER)

    p = parsers.add_parser(
        'safety',
        help='Run the safety vulnerability check'
    )
    p.add_argument('remainder', nargs=argparse.REMAINDER)

    return parser


@multi
def task(path, info, args):
    return args.phase


@method(task, 'virtualenv')
def task_virtualenv(path, info, args):
    with working_dir(path):
        changed = changed_state(info)
        if has_self_or_dependent_changes(changed):
            log.info('Running pyenv requirements in %s', path)

            env = tool_env()

            python_version = info.get('pyenv', {}).get('python-version')
            if python_version:
                env['PYTHON_VERSION'] = python_version

            with package_script("/tools/pyenv/requirements.sh") as f:
                with working_dir(path):
                    res = subprocess.call(
                        [f.name, path, info['project-name']],
                        env=env,
                    )
                    if res:
                        return FAIL
                    return OK
        else:
            log.info('not running pyenv requirements in %s', path)
            return OK


PYENV_URL = 'https://github.com/pyenv/pyenv.git'
PYENV_VENV_URL = 'https://github.com/yyuu/pyenv-virtualenv.git'


@method(task, 'install-pyenv')
def task_install_pyenv(path, info, args):
    pyenv_root = getenv('PYENV_ROOT') or expanduser("~/.pyenv")
    bash_profile = expanduser("~/.bash_profile")

    if not isdir(pyenv_root):
        Repo.clone_from(PYENV_URL, pyenv_root)

    if "pyenv init" not in slurp(bash_profile):
        with open(bash_profile, 'a') as f:
            f.write('export PYENV_ROOT="$HOME/.pyenv"\n')
            f.write('export PATH="$PYENV_ROOT/bin:$PATH"\n')
            f.write('eval "$(pyenv init -)"\n')
    return OK


@method(task, 'install-virtualenv')
def task_install_virtualenv(path, info, args):
    pyenv_root = getenv('PYENV_ROOT') or expanduser("~/.pyenv")
    dest = join(pyenv_root, "plugins/pyenv-virtualenv")
    bash_profile = expanduser("~/.bash_profile")
    if not isdir(dest):
        Repo.clone_from(PYENV_VENV_URL, dest)

    if "virtualenv-init" not in slurp(bash_profile):
        with open(expanduser("~/.bash_profile"), 'a') as f:
            f.write('eval "$(pyenv virtualenv-init -)"\n')
    return OK


@method(task, 'safety')
def task_safety(path, info, args):
    import subprocess
    from menhir.tool_utils import package_script

    with package_script("/tools/pyenv/safety.sh") as f:
        res = subprocess.call([f.name], env=tool_env())
        if res:
            return FAIL
    return OK



"""Tool to print menhir's configuration and state dict.

This can be used to inspect what the configuration a previous tool
produces, or the next tool sees.

It can output in python pprint, json or yaml formats.
"""
from __future__ import absolute_import

import argparse
import logging
import pprint as pp
import sys

import json

from ..fileutils import yaml_dumper
from ..tool import Tool
from ..tool_utils import OK, FAIL


log = logging.getLogger(__name__)


def tool():
    return Pprint()


class Pprint(Tool):

    def dir_info(tool, path, info):
        return {
            'project_recognised': False,
            'can_run': True,
        }

    def dependencies(tool, path):
        return []

    def arg_parser(tool, **kwargs):
        return parser(**kwargs)

    def execute_tool(tool, path, info, args):
        """Pprint the menhir config."""
        if args.format == 'pprint':
            pp.pprint(info, width=80)
        elif args.format == 'yaml':
            yaml = yaml_dumper()
            yaml.dump(info, sys.stdout)
        elif args.format == 'json':
            print(json.dumps(info, sort_keys=True, indent=4))
        else:
            log.error('Invalid output format: "%s', args.format)
            return FAIL
        return OK


def parser(**kwargs):
    """Configure arg parser for the tool options and arguments."""
    parser = argparse.ArgumentParser(
        description="Pretty print the menhir configuration.",
        **kwargs
    )
    parser.add_argument(
        "--format",
        default='yaml',
        choices=['json', 'pprint', 'yaml'],
        help="Output format.",
    )
    parser.add_argument('remainder', nargs=argparse.REMAINDER)
    return parser
